from flask import Flask, request
from flask_sqlalchemy import SQLAlchemy
from logging import getLogger
import textwrap

from ._version import __version__

_logger = getLogger('poff')

db = SQLAlchemy()

def create_app(config_file=None):
    app = Flask('poff')
    if config_file:
        app.config.from_pyfile(config_file)
    else:
        app.config.from_envvar('POFF_CONFIG_FILE')

    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

    db.init_app(app)

    from . import views

    app.register_blueprint(views.mod)

    @app.teardown_appcontext
    def teardown_appcontext(error):
        """ Commits the session if no error has occured, otherwise rollbacks. """
        if error is None:
            try:
                db.session.commit()
            except Exception: # pylint: disable=broad-except
                # Whoopsie! We can't
                db.session.rollback()
                _logger.exception('Exception happened during teardown commit.')
        else:
            # We have an exception, but it has probably already been handled by the modroriate handlers,
            # so just rollback the session and ignore the error
            db.session.rollback()
        db.session.remove()

    @app.errorhandler(500)
    def server_error(error):
        generic_error_handler(error)
        # Don't rely on any special mechanisms such as template loading or other resources,
        # as we don't know where the error is.
        return textwrap.dedent('''\
        <!doctype html>
        <title>Internal Server Error</title>
        <h1>Internal Server Error</h1>
        <p>
            I'm very sorry to report this, but something has gone wrong on the server.
            If you have administrator access, check the logs for more details, otherwise
            you should probably gently notify your sysadmin of the failure.
        ''')

    @app.context_processor
    def default_context():
        return {
            'version': __version__,
        }

    return app

def generic_error_handler(exception):
    """ Log exception to the standard logger. """
    log_msg = textwrap.dedent("""Error occured.
        Path:                 %s
        Params:               %s
        HTTP Method:          %s
        Client IP Address:    %s
        User Agent:           %s
        User Platform:        %s
        User Browser:         %s
        User Browser Version: %s
        HTTP Headers:         %s
        Exception:            %s
        """ % (
            request.path,
            request.values,
            request.method,
            request.remote_addr,
            request.user_agent.string,
            request.user_agent.platform,
            request.user_agent.browser,
            request.user_agent.version,
            request.headers,
            exception
        )
    )
    _logger.exception(log_msg)



from poff import create_app, db

from functools import partial
import os
import tempfile
import unittest


class DBTestCase(unittest.TestCase):

    # used to create the assertXXX helpers
    _assert_helpers = (
        200,
        201,
        400,
        403,
        404,
    )

    def _set_up(self):
        self.config_file = tempfile.NamedTemporaryFile(delete=False)
        self.config_file.write('\n'.join([
            'SQLALCHEMY_DATABASE_URI = "sqlite://"',
            'SECRET_KEY = "testkey"',
#            'TESTING = True',
            'WTF_CSRF_ENABLED = False',
        ]).encode('utf-8'))
        self.config_file.close()
        self.app = create_app(self.config_file.name)
        with self.app.app_context():
            db.create_all()
        self.client = self.app.test_client()
        for code in DBTestCase._assert_helpers:
            setattr(self, 'assert%d' % code, partial(self.assert_status_code, code=code))


    def assert_status_code(self, response, code):
        self.assertEqual(response.status_code, code)


    def _tear_down(self):
        os.remove(self.config_file.name)


    def __call__(self, *args, **kwargs):
        self._set_up()
        if hasattr(self, 'set_up'):
            self.set_up()
        super(DBTestCase, self).__call__(*args, **kwargs)
        if hasattr(self, 'tear_down'):
            self.tear_down()
        self._tear_down()


    def add_objects(self, *args):
        ids = []
        with self.app.test_request_context():
            for obj in args:
                db.session.add(obj)
            db.session.commit()
            for obj in args:
                ids.append(obj.id)
        return ids


    def assertForbidden(self, response):
        self.assert403(response)



from . import DBTestCase
from poff.models import Domain, Record, DomainMeta, TsigKey

import datetime
import re

class DomainTest(DBTestCase):

    def set_up(self):
        domain = Domain(name='example.com')
        soa_record = Record(name='example.com', type='SOA', content='example.com hostmaster.example.com 1970010101', domain=domain)
        record = Record(name='www.example.com', type='A', content='127.0.0.1', domain=domain)
        self.domain_id, _, _ = self.add_objects(domain, record, soa_record)


    def test_create_domain(self):
        data = {
            'name': 'test.com',
        }
        response = self.client.post('/domains', data=data, follow_redirects=True)
        self.assert200(response)
        with self.app.app_context():
            domains = Domain.query.filter(Domain.id != self.domain_id).all()
            self.assertEqual(len(domains), 1)

            # SOA record should have been created
            soa_record = Record.query.filter(Record.type=='SOA', Record.name=='test.com').one()
            today = datetime.datetime.now()
            self.assertEqual(soa_record.serial, today.strftime('%Y%m%d00'))

            # No-mail SPF record should have been created
            spf_record = Record.query.filter(Record.type=='TXT',
                Record.name=='test.com').one()
            self.assertEqual(spf_record.content, 'v=spf1 -all')

            # SOA-EDIT should be set
            soa_edit_record = DomainMeta.query.filter_by(domain=domains[0], kind='SOA-EDIT').first()
            self.assertEqual(soa_edit_record.content, 'INCEPTION-INCREMENT')

            # Should add NSEC3 parameters
            nsec3params = DomainMeta.query.filter_by(domain=domains[0], kind='NSEC3PARAMS').first()
            self.assertTrue(re.match(r'^1 0 1 [a-f0-9]{32}$', nsec3params.content))

            # Should set NSEC3 to narrow mode
            nsec3narrow = DomainMeta.query.filter_by(domain=domains[0], kind='NSEC3NARROW').first()
            self.assertEqual(nsec3narrow.content, '1')


    def test_mname_validation(self):
        invalid_mnames = (
            'with spaces.example.com',
        )
        for invalid_mname in invalid_mnames:
            self.assert400(self.client.post('/domains', data={
                'name': 'example.com',
                'mname': invalid_mname,
            }))

        valid_mnames = (
            'example.com',
            'ns.example.com',
            'otherdomain.tld',
        )
        for valid_mname in valid_mnames:
            self.assert200(self.client.post('/domains', data={
                'name': 'example.com',
                'mname': valid_mname,
            }, follow_redirects=True))


    def test_rname_validation(self):
        invalid_rnames = (
            'with spaces@example.com',
            'notld',
        )
        for invalid_rname in invalid_rnames:
            self.assert400(self.client.post('/domains', data={
                'name': 'example.com',
                'rname': invalid_rname,
            }))

        valid_rnames = (
            'me@exammple.com',
            'me.example.com',
            'some.gal123@gmail.co.uk',
        )
        for valid_rname in valid_rnames:
            self.assert200(self.client.post('/domains', data={
                'name': 'example.com',
                'rname': valid_rname,
            }, follow_redirects=True))


    def test_modify_mname(self):
        response = self.client.post('/domains/%d' % self.domain_id, data={
            'mname': 'ns.example.com',
            'rname': 'john.doe@example.com',
        }, follow_redirects=True)
        self.assert200(response)

        with self.app.app_context():
            soa_record = Record.query.filter(Record.type=='SOA').one()
            mname, rname, serial = soa_record.content.split(' ')[:3]
            self.assertEqual(mname, 'ns.example.com')
            self.assertEqual(rname, 'john\.doe.example.com')
            self.assertNotEqual(serial, '1970010101')


    def test_modify_mname_invalid(self):
        self.assert400(self.client.post('/domains/%d' % self.domain_id, data={
            'mname': 'ns with spaces.com',
        }))


    def test_modify_rname_invalid(self):
        self.assert400(self.client.post('/domains/%d' % self.domain_id, data={
            'rname': 'with spaces@gmail.com',
        }))


    def test_create_tsigkey(self):
        data = {
            'name': 'testkey',
        }
        response = self.client.post('/domains/%d/tsigkeys' % self.domain_id, data=data,
            follow_redirects=True)
        self.assert200(response)

        with self.app.app_context():
            tsigkeys = TsigKey.query.all()
            self.assertEqual(len(tsigkeys), 1)

            domain_tsigkeys = DomainMeta.query\
                .filter_by(domain_id=self.domain_id, kind='TSIG-ALLOW-DNSUPDATE').all()
            self.assertEqual(len(domain_tsigkeys), 1)
            self.assertEqual(domain_tsigkeys[0].content, 'testkey')


    def test_delete_tsigkey(self):
        tsigkey = TsigKey(name='testkey')
        tsigmeta = DomainMeta(domain_id=self.domain_id, kind='TSIG-ALLOW-DNSUPDATE',
            content='testkey')
        self.add_objects(tsigkey, tsigmeta)

        response = self.client.delete('/domains/%d/tsigkeys/testkey' % self.domain_id,
            follow_redirects=True)

        self.assert200(response)
        with self.app.app_context():
            self.assertEqual(len(TsigKey.query.all()), 0)
            self.assertEqual(len(DomainMeta.query.all()), 0)


    def test_delete_tsigkey_over_post(self):
        tsigkey = TsigKey(name='testkey')
        tsigmeta = DomainMeta(domain_id=self.domain_id, kind='TSIG-ALLOW-DNSUPDATE',
            content='testkey')
        self.add_objects(tsigkey, tsigmeta)

        response = self.client.post('/domains/%d/tsigkeys/testkey' % self.domain_id,
            data={'_method': 'DELETE'}, follow_redirects=True)

        self.assert200(response)
        with self.app.app_context():
            self.assertEqual(len(TsigKey.query.all()), 0)
            self.assertEqual(len(DomainMeta.query.all()), 0)


    def test_delete_nonexisting_tsigkey(self):
        response = self.client.delete('/domains/%d/tsigkeys/foobar' % self.domain_id)
        self.assert404(response)


    def test_create_invalid_domain(self):
        data = {
            'name': '',
        }
        response = self.client.post('/domains', data=data, follow_redirects=True)
        self.assert400(response)


    def test_delete_domain(self):
        response = self.client.delete('/domains/%d' % self.domain_id, follow_redirects=True)
        self.assert200(response)
        with self.app.app_context():
            domains = Domain.query.all()
            self.assertEqual(len(domains), 0)

            # associated records should have been cascade deleted
            records = Record.query.all()
            self.assertEqual(len(records), 0)


    def test_update_soa_serial(self):
        soa_record = Record(type='SOA', content='x y 2014010100')
        new_date = datetime.date(2014, 1, 2)
        soa_record.update_serial(new_date)
        self.assertEqual(soa_record.serial, '2014010200')
        soa_record.update_serial(new_date)
        self.assertEqual(soa_record.serial, '2014010201')
        soa_record.update_serial()
        today = datetime.datetime.now()
        self.assertEqual(soa_record.serial, today.strftime('%Y%m%d00'))



from . import DBTestCase
from poff.models import Domain, DynDNSClient, Record

class DynDNSTest(DBTestCase):

    def set_up(self):
        domain = Domain(name='test.com')
        soa_record = Record(name='test.com', type='SOA', content='x y 2014010100', domain=domain)
        record = Record(name='www.test.com', type='A', content='127.0.0.1', domain=domain)
        client = DynDNSClient(record=record)
        self.client_key = client.printable_key
        _, self.soa_id, self.record_id, self.client_id = self.add_objects(domain, soa_record,
            record, client)


    def test_create_client(self):
        response = self.client.post('/records/%d/new-dyndns-client' % self.record_id,
            follow_redirects=True)
        self.assert200(response)
        with self.app.app_context():
            clients = DynDNSClient.query.all()
            self.assertEqual(len(clients), 2)


    def test_update_record(self):
        data = {
            'record': 'www.test.com',
            'key': self.client_key,
        }
        headers = {
            'X-Forwarded-For': '1.2.3.4',
        }
        response = self.client.post('/update-record', data=data,
            environ_overrides={'REMOTE_ADDR':'127.0.0.1'}, headers=headers,
        )
        self.assert201(response)

        with self.app.app_context():
            # soa serial number should have been updated
            new_serial = Record.query.get(self.soa_id).serial
            self.assertNotEqual(new_serial, '2014010100')

            record = Record.query.get(self.record_id)
            self.assertEqual(record.content, '1.2.3.4')

        response = self.client.post('/update-record', data=data,
            environ_overrides={'REMOTE_ADDR':'127.0.0.1'}, headers=headers)

        # Shouldn't change anything with a second update
        with self.app.app_context():
            self.assertEqual(Record.query.get(self.soa_id).serial, new_serial)


    def test_update_tunneled_ipv4_record(self):
        origin = '::ffff:10.10.10.10'
        data = {
            'record': 'www.test.com',
            'key': self.client_key,
        }
        headers = {'X-Forwarded-For': origin}
        response = self.client.post('/update-record', data=data, follow_redirects=True,
            environ_overrides={'REMOTE_ADDR': '127.0.0.1'}, headers=headers)
        self.assert201(response)

        # Test that correct IP was set
        with self.app.app_context():
            self.assertEqual(Record.query.get(self.record_id).content, '10.10.10.10')


    def test_update_record_invalid_key(self):
        data = {
            'record': 'www.test.com',
            'key': 'hopefully invalid',
        }
        response = self.client.post('/update-record', data=data)
        self.assertForbidden(response)
        with self.app.app_context():
            # soa serial number should NOT have been updated
            self.assertEqual(Record.query.get(self.soa_id).serial, '2014010100')


    def test_update_record_invalid_record(self):
        data = {
            'record': 'foo.test.com',
            'key': self.client_key,
        }
        response = self.client.post('/update-record', data=data)
        self.assertEqual(response.status_code, 404)
        with self.app.app_context():
            # soa serial number should NOT have been updated
            self.assertEqual(Record.query.get(self.soa_id).serial, '2014010100')


    def test_delete_client(self):
        self.assert200(self.client.delete('/dyndns-clients/%d' % self.client_id, follow_redirects=True))
        with self.app.app_context():
            clients = DynDNSClient.query.all()
            self.assertEqual(len(clients), 0)


    def test_rekey_record(self):
        response = self.client.post('/records/%d/rekey' % self.record_id, follow_redirects=True)
        self.assert200(response)
        with self.app.app_context():
            new_key = Record.query.get(self.record_id).dyndns_client.printable_key
            self.assertNotEqual(new_key, self.client_key)


    def test_rekey_invalid(self):
        response = self.client.post('/records/0101/rekey')
        self.assert404(response)



from . import DBTestCase

from poff.models import Domain

class DispatchTest(DBTestCase):

    def set_up(self):
        domain = Domain(name='test.com')
        self.domain_id, = self.add_objects(domain)


    def test_post_rewrite(self):
        data = {'_method': 'DELETE'}
        #self.assert200(self.client.delete('/domains/%d' % self.domain_id, follow_redirects=True))
        response = self.client.post('/domains/%d' % self.domain_id, data=data, follow_redirects=True)
        self.assert200(response)



import math
import string
import sys

ALPHABET = string.digits + string.ascii_letters
PY2 = sys.version_info[0] == 2

if PY2:
    def ord_byte(char):
        return ord(char)
else:
    def ord_byte(byte):
        return byte


def ceildiv(dividend, divisor):
    ''' integer ceiling division '''
    return (dividend + divisor - 1) // divisor


def calc_chunklen(alph_len):
    '''
    computes the ideal conversion ratio for the given alphabet.
    A ratio is considered ideal when the number of bits in one output
    encoding chunk that don't add up to one input encoding chunk is minimal.
    '''
    binlen, enclen = min([
                          (i, i*8 / math.log(alph_len, 2))
                          for i in range(1, 7)
                         ], key=lambda k: k[1] % 1)

    return binlen, int(enclen)


def encode(digest):
    chunklen = calc_chunklen(len(ALPHABET))
    nchunks = ceildiv(len(digest), chunklen[0])
    binstr = digest.ljust(nchunks * chunklen[0], b'\0')

    return ''.join([
            _encode_chunk(binstr, i) for i in range(0, nchunks)
        ])

def _encode_chunk(data, index):
    '''
    gets a chunk from the input data, converts it to a number and
    encodes that number
    '''
    chunk = _get_chunk(data, index)
    return _encode_long(_chunk_to_long(chunk))


def _encode_long(val):
    '''
    encodes an integer of 8*chunklen[0] bits using the specified
    alphabet
    '''
    chunklen = calc_chunklen(len(ALPHABET))
    return ''.join([
            ALPHABET[(val//len(ALPHABET)**i) % len(ALPHABET)]
            for i in reversed(range(chunklen[1]))
        ])


def _chunk_to_long(chunk):
    '''
    parses a chunk of bytes to integer using big-endian representation
    '''
    chunklen = calc_chunklen(len(ALPHABET))
    return sum([
            256**(chunklen[0]-1-i) * ord_byte(chunk[i])
            for i in range(chunklen[0])
        ])


def _get_chunk(data, index):
    '''
    partition the data into chunks and retrieve the chunk at the given index
    '''
    chunklen = calc_chunklen(len(ALPHABET))
    return data[index*chunklen[0]:(index+1)*chunklen[0]]



from . import DBTestCase
from poff.models import Domain, Record

class RecordTest(DBTestCase):

    def set_up(self):
        domain = Domain(name='test.com')
        soa_record = Record(type='SOA', content='x y 2014010100', name='test.com', domain=domain)
        record = Record(name='www.test.com', type='A', content='127.0.0.1', domain=domain)
        self.domain_id, self.soa_id, self.record_id = self.add_objects(domain, soa_record, record)


    def test_create_record(self):
        data = {
            'name': 'ftp.test.com',
            'type': 'A',
            'content': '127.0.0.1',
        }
        response = self.client.post('/domains/%d/new_record' % self.domain_id, data=data,
            follow_redirects=True)
        self.assert200(response)
        with self.app.app_context():
            records = Record.query.all()
            self.assertEqual(len(records), 3)

            # soa serial number should have been updated
            self.assertNotEqual(Record.query.get(self.soa_id).serial, '2014010100')


    def test_create_mx_record_default_prio(self):
        data = {
            'name': 'mx.test.com',
            'type': 'MX',
            'content': '127.0.0.1',
        }
        response = self.client.post('/domains/%d/new_record' % self.domain_id, data=data,
            follow_redirects=True)
        self.assert200(response)
        with self.app.app_context():
            records = Record.query.filter_by(type='MX').all()
            self.assertEqual(len(records), 1)
            self.assertEqual(records[0].prio, 0)


    def test_invalid_create_record(self):
        data = {
            'name': 'ftp.test.com',
            'type': 'BOGUS',
            'content': 'lols',
        }
        response = self.client.post('/domains/%d/new_record' % self.domain_id, data=data)
        self.assert400(response)


    def test_modify_record(self):
        data = {
            'name': 'www.test.com',
            'type': 'A',
            'content': '127.0.0.2',
        }
        response = self.client.post('/records/%d' % self.record_id, data=data,
            follow_redirects=True)
        self.assert200(response)
        with self.app.app_context():
            record = Record.query.get(self.record_id)
            self.assertEqual(record.content, '127.0.0.2')


    def test_delete_record(self):
        response = self.client.delete('/records/%d' % self.record_id, follow_redirects=True)
        self.assert200(response)
        with self.app.app_context():
            records = Record.query.all()
            self.assertEqual(len(records), 1)

            # soa serial number should have been updated
            self.assertNotEqual(Record.query.get(self.soa_id).serial, '2014010100')


    def test_delete_html(self):
        data = {
            '_method': 'DELETE',
        }
        self.assert200(self.client.post('/records/%d' % self.record_id, data=data,
            follow_redirects=True))


    def test_record_sort_order(self):
        subsubdomains = ['a', 'z', 'b']
        for num, subsubdomain in enumerate(subsubdomains):
            data = {
                'name': '%s.subdomain.test.com' % subsubdomain,
                'type': 'A',
                'content': '127.0.0.%d' % num
            }
            response = self.client.post('/domains/%d/new_record' % self.domain_id, data=data,
                follow_redirects=True)
            self.assert200(response)
        subsubdomains.sort()
        with self.app.app_context():
            domain = Domain.query.get(self.domain_id)
            num = 0
            for record in domain.records:
                if not 'subdomain' in record.name:
                    continue
                self.assertEqual(record.name, subsubdomains[num] + '.subdomain.test.com')
                num += 1



from . import db, base62
from .models import (Domain, DomainForm, DynDNSClient, Record, RecordForm, DomainMeta, TsigKey,
    TsigKeyForm)

from flask import abort, redirect, render_template, flash, request, Blueprint
from flask.views import MethodView
from logging import getLogger
from itsdangerous.signer import constant_time_compare
import os

_logger = getLogger('poff.views')

mod = Blueprint('views', __name__)


@mod.context_processor
def default_context():
    domains = Domain.query.order_by(Domain.name).all()
    return {
        'domains': domains,
        'tsigkeyform': TsigKeyForm(),
        'domainform': DomainForm(),
        'recordform': RecordForm(),
    }


@mod.route('/')
def main():
    return render_template('domains.html')


@mod.route('/domains', methods=['POST'])
def domains():
    form = DomainForm()
    if form.validate_on_submit():
        domain = Domain()
        mname = form.mname.data or form.name.data
        rname = form.rname.data or 'hostmaster.' + form.name.data
        del form['mname']
        del form['rname']
        form.populate_obj(domain)

        soa_record = Record(content='%(mname)s %(rname)s 1970010100' % {
                'mname': mname,
                'rname': rname,
            },
            domain=domain,
            type='SOA',
            name=domain.name,
        )
        soa_record.update_serial()

        spf_record = Record(content='v=spf1 -all', domain=domain, type='TXT',
            name=domain.name)

        db.session.add(domain)
        db.session.add(soa_record)
        db.session.add(spf_record)
        db.session.add(DomainMeta(domain=domain, kind='SOA-EDIT', content='INCEPTION-INCREMENT'))
        db.session.add(DomainMeta(domain=domain, kind='NSEC3NARROW', content='1'))
        nsec3params = '1 0 1 %s' % os.urandom(16).hex()
        db.session.add(DomainMeta(domain=domain, kind='NSEC3PARAMS', content=nsec3params))
        _logger.info('New domain saved: %s', domain.name)
        flash('New domain added successfully!', 'success')
    else:
        flash('Failed to validate new domain, check the errors in the form below', 'error')
        _logger.debug('New domain failed form validation.')
        return render_template('domains.html', domainform=form), 400
    return redirect('/')


class MethodOverrideView(MethodView):
    """ Helper class that rewrites POST requests with a `_method` param to the given HTTP method.
    """

    allowed_methods = frozenset([
        'GET',
        'HEAD',
        'POST',
        'DELETE',
        'PUT',
        'PATCH',
        'OPTIONS'
    ])

    # Make sure POST requests always are allowed to this view
    methods = ['POST']

    def dispatch_request(self, *args, **kwargs):
        """ Find the target HTTP method and execute it. """
        meth = getattr(self, request.method.lower(), None)

        # if the request method is HEAD and we don't have a handler for it retry with GET
        if meth is None and request.method == 'HEAD':
            meth = getattr(self, 'get', None)

        # If an override is desired, load that if we have it
        method_override = request.form.get('_method', '').upper()
        if method_override in MethodOverrideView.allowed_methods and hasattr(self, method_override.lower()):
            meth = getattr(self, method_override.lower())

        assert meth is not None, 'Unimplemented method %r' % request.method
        return meth(*args, **kwargs)


class DomainView(MethodOverrideView):

    def post(self, domain_id):
        domain = Domain.query.get_or_404(domain_id)
        form = DomainForm()

        # This isn't modifiable, and we're re-using the form because we're lazy
        del form['name']

        if not form.validate_on_submit():
            _logger.info('Failed to validate domain form')
            abort(400)

        form.populate_obj(domain)
        domain.update_soa()
        return redirect('/')


    def delete(self, domain_id):
        domain = Domain.query.get_or_404(domain_id)
        db.session.delete(domain)
        _logger.info("Deleting domain %s", domain.name)
        flash('Domain %s deleted successfully' % domain.name, 'info')
        return redirect('/')


class RecordView(MethodOverrideView):

    def post(self, record_id):
        record = Record.query.get_or_404(record_id)
        form = RecordForm()
        if form.validate_on_submit():
            form.populate_obj(record)
            domain = record.domain
            domain.update_soa()
            _logger.info('Record %s modified', record.name)
            flash('Record successfully modified.', 'success')
            return redirect('/')
        else:
            _logger.info('Record modification failed form validation: %s', form.errors)
            flash('Failed to validate record modifications', 'warning')
            context = {
                'form_errors': form.errors,
                'recordform': form,
            }
            return render_template('domains.html', **context), 400


    def delete(self, record_id):
        record = Record.query.get_or_404(record_id)
        record.domain.update_soa()
        db.session.delete(record)
        _logger.info("Deleting record %s", record.name)
        flash('Record %s deleted successfully.' % record.name, 'success')
        return redirect('/')


@mod.route('/domains/<int:domain_id>/new_record', methods=['POST'])
def new_record(domain_id):
    domain = Domain.query.get_or_404(domain_id)
    form = RecordForm()
    if form.validate_on_submit():
        record = Record()
        record.domain = domain
        form.populate_obj(record)

        # Ensure MX records always has a prio
        if record.type == 'MX':
            record.prio = record.prio or 0

        domain.update_soa()
        db.session.add(record)
        _logger.info('New record saved: %s', record.name)
        flash('New record saved successfully!', 'success')
    else:
        _logger.debug('Record failed form validation')
        flash('Failed to validate new record, check the errors in the form below!', 'warning')
        return render_template('domains.html', recordform=form), 400
    return redirect('/')


@mod.route('/domains/<int:domain_id>/tsigkeys', methods=['POST'])
def domain_tsigkeys(domain_id):
    domain = Domain.query.get_or_404(domain_id)
    form = TsigKeyForm()

    if not form.validate_on_submit():
        _logger.debug('TsigKey failed form validation')
        flash('Failed to validate new DynDNS key', 'warning')
        return render_template('domains.html', tsigkeyform=form), 400

    tsigkey = TsigKey()
    form.populate_obj(tsigkey)

    domain_tsigkey = DomainMeta(domain=domain, kind='TSIG-ALLOW-DNSUPDATE',
        content=tsigkey.name)

    db.session.add(tsigkey)
    db.session.add(domain_tsigkey)

    _logger.info('Added tsigkey for for "%s"', domain.name)
    flash('New dyndns key added successfully!', 'success')
    return redirect('/')


class TsigKeyView(MethodOverrideView):

    def delete(self, domain_id, tsig_name):
        domain = Domain.query.get_or_404(domain_id)
        tsigkey = TsigKey.query.filter_by(name=tsig_name).first() or abort(404)
        tsig_meta = DomainMeta.query.filter_by(domain_id=domain_id, kind='TSIG-ALLOW-DNSUPDATE',
            content=tsig_name).first()

        db.session.delete(tsigkey)
        db.session.delete(tsig_meta)

        _logger.info('DynDNS key %s deleted', tsigkey.name)
        flash('DynDNS key %s deleted' % tsigkey.name, 'success')
        return redirect('/')


@mod.route('/records/<int:record_id>/new-dyndns-client', methods=['POST'])
def new_dyndns_client(record_id):
    record = Record.query.get_or_404(record_id)
    client = DynDNSClient(record=record)
    db.session.add(client)
    _logger.info('New DynDNS client created for record %s', record.name)
    flash('New DynDNS client created!', 'success')
    return redirect('/')


@mod.route('/records/<int:record_id>/rekey', methods=['POST'])
def rekey_dyndns_record(record_id):
    record = Record.query.get(record_id)
    if not record or not record.dyndns_client:
        abort(404)
    dyndns_client = record.dyndns_client
    dyndns_client.set_new_key()
    return redirect('/')


@mod.route('/update-record', methods=['POST'])
def update_record():
    record_name = request.form.get('record')
    record = Record.query.filter_by(name=record_name, type='A').first()
    if not record or not record.dyndns_client:
        abort(404)
    submitted_key = str(request.form.get('key', ''))
    record_key = base62.encode(record.dyndns_client.key)
    if constant_time_compare(submitted_key, record_key):
        origin_ip = request.access_route[0]
        if origin_ip.startswith('::ffff:'):
            origin_ip = origin_ip[len('::ffff:'):]
        if record.content != origin_ip:
            _logger.info('Updating record %s to %s', record.name, origin_ip)
            flash('Successfully updated record to new IP: %s' % origin_ip, 'success')
            record.content = origin_ip
            record.domain.update_soa()
            return '', 201
        else:
            flash('Still on the same IP, no change applied', 'success')
            return ''
    else:
        _logger.warning('Bad auth for trying to update record %s', record.name)
        abort(403)


class DynDNSClientView(MethodOverrideView):

    def delete(self, client_id):
        client = DynDNSClient.query.get_or_404(client_id)
        db.session.delete(client)
        _logger.info('DynDNS client for record %s deleted.', client.record.name)
        flash('DynDNS client deleted!')
        return redirect('/')


mod.add_url_rule('/domains/<int:domain_id>', view_func=DomainView.as_view('domain_details'))
mod.add_url_rule('/records/<int:record_id>', view_func=RecordView.as_view('record_details'))
mod.add_url_rule('/dyndns-clients/<int:client_id>',
    view_func=DynDNSClientView.as_view('client_details'), methods=['POST', 'DELETE'])
mod.add_url_rule('/domains/<int:domain_id>/tsigkeys/<tsig_name>',
    view_func=TsigKeyView.as_view('tsigkey_details'), methods=['POST', 'DELETE'])



from . import db, base62

from flask import Markup
from flask_wtf import Form
from sqlalchemy.ext.hybrid import hybrid_property
from wtforms.fields import HiddenField, TextField
from wtforms.validators import Regexp, Optional
from wtforms_alchemy import model_form_factory
import base64
import datetime
import os

# Types supported by PowerDNS, see http://doc.powerdns.com/html/types.html
_RECORD_TYPES = (
    'A',
    'AAAA',
    'AFSDB',
    'ALIAS',
    'CAA',
    'CDNSKEY',
    'CDS',
    'CERT',
    'CNAME',
    'DNSKEY',
    'DS',
    'HINFO',
    'KEY',
    'LOC',
    'MX',
    'NAPTR',
    'NS',
    'NSEC',
    'PTR',
    'RP',
    'RRSIG',
    'SOA',
    'SPF',
    'SSHFP',
    'SRV',
    'TXT',
)

_TSIG_ALGORITHMS = (
    'hmac-sha256',
    'hmac-md5',
    'hmac-sha1',
    'hmac-sha224',
    'hmac-sha384',
    'hmac-sha512',
)


class Domain(db.Model):
    __tablename__ = 'domains'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), nullable=False)
    master = db.Column(db.String(128))
    last_check = db.Column(db.Integer)
    type = db.Column(db.String(6), nullable=False, default='MASTER')
    notified_serial = db.Column(db.Integer)
    account = db.Column(db.String(40))

    @property
    def records(self):
        """ Sort records such that subdomains are grouped together. """
        return sorted(self._records, key=lambda r: '.'.join(reversed(r.name.split('.'))))


    @property
    def soa_record(self):
        return Record.query.filter(Record.type=='SOA', Record.domain_id==self.id).one()


    def update_soa(self):
        """ Update the serial number of the SOA associated with this domain. """
        self.soa_record.update_serial()


    @property
    def tsigkeys(self):
        keys = TsigKey.query.join(DomainMeta, TsigKey.name == DomainMeta.content)\
            .filter(DomainMeta.kind=='TSIG-ALLOW-DNSUPDATE')\
            .filter(DomainMeta.domain_id==self.id)\
            .all()
        return keys


    @hybrid_property
    def mname(self):
        return self.soa_record.content.split(' ')[0]


    @mname.setter
    def set_mname(self, value):
        soa_record = self.soa_record
        parts = soa_record.content.split(' ')
        parts[0] = value
        soa_record.content = ' '.join(parts)


    @hybrid_property
    def rname(self):
        return self.soa_record.content.split(' ')[1].replace('.', '@', 1)


    @rname.setter
    def set_rname(self, value):
        soa_record = self.soa_record
        parts = soa_record.content.split(' ')
        label, domain = value.split('@', 1)
        label = label.replace('.', '\.')
        parts[1] = '%s.%s' % (label, domain)
        soa_record.content = ' '.join(parts)


    @property
    def form(self):
        return DomainForm(obj=self)


class DomainMeta(db.Model):
    __tablename__ = 'domainmetadata'
    id = db.Column(db.Integer, primary_key=True)
    domain = db.relationship('Domain', backref=db.backref('_metadata', lazy='dynamic', cascade='delete'))
    domain_id = db.Column(db.Integer, db.ForeignKey('domains.id'))
    kind = db.Column(db.String(16))
    content = db.Column(db.Text())


class TsigKey(db.Model):
    __tablename__ = 'tsigkeys'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255))
    algorithm = db.Column(db.String(50), info={
        'choices': [(t, t) for t in _TSIG_ALGORITHMS],
    }, default='hmac-sha256')
    secret = db.Column(db.String(255))

    def __init__(self, **kwargs):
        if not 'secret' in kwargs:
            kwargs['secret'] = base64.b64encode(os.urandom(16))

        super(TsigKey, self).__init__(**kwargs)


class Record(db.Model):
    __tablename__ = 'records'
    id = db.Column(db.Integer, primary_key=True)
    domain_id = db.Column(db.Integer, db.ForeignKey('domains.id'))
    name = db.Column(db.String(255), nullable=False)
    type = db.Column(db.String(10), info={
        'choices': [(t, t) for t in _RECORD_TYPES],
    })
    content = db.Column(db.String(65535))
    ttl = db.Column(db.Integer, default=3600)
    prio = db.Column(db.Integer)
    change_date = db.Column(db.Integer)
    disabled = db.Column(db.Boolean, default=False)
    domain = db.relationship('Domain', backref=db.backref('_records', lazy='dynamic', cascade='all,delete'))

    @property
    def serial(self):
        """ Get the serial number of a SOA record. """
        assert self.type == 'SOA', 'tried to get serial from non-SOA record'
        parts = self.content.split()
        assert len(parts) >= 2, "SOA record didn't have serial number"
        serial_number = parts[2]
        return serial_number


    @serial.setter
    def serial(self, value):
        assert self.type == 'SOA', 'tried to set serial number of non-SOA record'
        parts = self.content.split()
        assert len(parts) >= 2, 'Malformed SOA record, no serial number set'
        # content is formatted 'hostname admin_email serial_number ..'
        parts[2] = value
        self.content = ' '.join(parts)


    def update_serial(self, date=None):
        """ Update the serial number of the SOA record, assuming the given date.

        Will set the serial number according to the YYYYMMDDss convention, where ss is a number
        incremented once per change per day, and reset every day.
        """
        assert self.type == 'SOA', 'tried to update serial number of non-SOA record'
        if not date:
            date = datetime.datetime.now().date()
        current_serial = self.serial
        serial_date = datetime.datetime.strptime(current_serial[:8], '%Y%m%d')
        change_num = int(current_serial[8:])
        if serial_date.date() == date:
            change_num += 1
        else:
            change_num = 0
        self.serial = '%s%02d' % (date.strftime('%Y%m%d'), change_num)


class DynDNSClient(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    record_id = db.Column(db.Integer, db.ForeignKey('records.id'))
    key = db.Column(db.LargeBinary(64), nullable=False)
    record = db.relationship('Record', backref=db.backref('dyndns_client', uselist=False))


    def __init__(self, *args, **kwargs):
        self.set_new_key()
        super(DynDNSClient, self).__init__(*args, **kwargs)


    def set_new_key(self):
        self.key = os.urandom(30)


    @property
    def printable_key(self):
        return base62.encode(self.key)


class _PrintableForm(model_form_factory(Form)):

    def render(self):
        fields = []
        for f in self:
            if isinstance(f, HiddenField):
                fields.append('<input type="hidden" name="%(name)s" value="%(value)s">' % {
                    'name': f.name,
                    'value': f._value(),
                })
            else:
                fields.append('%s: %s' % (f.label, f()))
        return Markup('\n'.join(fields))


class DomainForm(_PrintableForm):
    class Meta:
        model = Domain
        only = (
            'name',
        )

    mname = TextField(label='Master nameserver', validators=(
        Optional(),
        Regexp(r'^[\w\.]+$'),
    ))
    rname = TextField(label='Contact email', validators=(
        Optional(),
        # Simple email subset that also allows pre-formatted domain records
        Regexp(r'^[\w_.+-]+(@|\.)[\w._-]+'),
    ))


class RecordForm(_PrintableForm):
    class Meta:
        model = Record
        only = (
            'name',
            'type',
            'content',
            'ttl',
            'prio',
        )


class TsigKeyForm(_PrintableForm):
    class Meta:
        model = TsigKey
        only = (
            'name',
            'algorithm',
        )



from  . import create_app, db
from .models import DynDNSClient

import argparse
import logging.config
import yaml
from sqlalchemy.schema import CreateTable


_CONFIG_FILE_PARSER = argparse.ArgumentParser(add_help=False)

_CONFIG_FILE_PARSER.add_argument('-c', '--config-file',
    metavar='<config-file>',
    help='Config file to use. If none is given, will load from the envvar POFF_CONFIG_FILE.',
)


def cli_entry():
    """ CLI entrypoint. """
    parser = argparse.ArgumentParser(prog='poff')

    subparser = parser.add_subparsers(title='action',
        help='Action to be performed',
    )

    add_init_parser(subparser)
    add_serve_parser(subparser)

    args = parser.parse_args()
    args.target(args)


def add_serve_parser(subparser):
    """ Add the parser for the `serve` command. """
    parser = subparser.add_parser('serve',
        help='Run the webserver',
        parents=[_CONFIG_FILE_PARSER],
    )
    parser.add_argument('-H', '--host',
        metavar='<host>',
        default='127.0.0.1',
        help='Which address to listen to. Default: %(default)s',
    )
    parser.add_argument('-p', '--port',
        metavar='<port>',
        type=int,
        default=5353,
        help='Which port to bind to. Default: %(default)s',
    )
    parser.add_argument('-d', '--debug',
        action='store_true',
        default=False,
        help='Show debug inforamtion on errors. DO NOT RUN WITH THIS OPTION IN PRODUCTION!',
    )
    parser.add_argument('-l', '--log-config',
        metavar='<log-config-file>',
        help='Path to a YAML file to load logging config from.',
    )

    parser.set_defaults(target=serve)


def add_init_parser(subparser):
    """ Add the `init` command parser. """
    parser = subparser.add_parser('init',
        help='Create the database tables',
        parents=[_CONFIG_FILE_PARSER],
    )
    parser.add_argument('-p', '--print',
        action='store_true',
        help='Print the table creation SQL instead of executing it. The SQL assumes that the ' +
        'rest of the tables has already been created.',
    )
    parser.set_defaults(target=init)


def serve(args):
    """ Run the webserver. """
    _init_logging(args)
    app = create_app(config_file=args.config_file)
    app.run(host=args.host, port=args.port, debug=args.debug)


def init(args):
    """ Initialize the database tables. """
    app = create_app(config_file=args.config_file)
    with app.app_context():
        if getattr(args, 'print'):
            print(CreateTable(DynDNSClient.__table__).compile(db.engine))
        else:
            db.create_all()


def _init_logging(args):
    if args.log_config:
        with open(args.log_config) as log_config_fh:
            logging_config = yaml.load(log_config_fh)
            logging.config.dictConfig(logging_config)
    else:
        log_format = '%(asctime)s %(levelname)-10s %(name)s %(message)s'
        level = logging.DEBUG if args.debug else logging.INFO
        logging.basicConfig(format=log_format, level=level)
